package Msn::Bot;

# $Id: Bot.pm 29 2004-03-21 07:12:54Z tea $

use strict;
use warnings;
use Carp;

use Msn;
use Msn::Client;
use Msn::Message;
use Msn::Connection;
use Msn::Authentication;
use Msn::Message::MsnObject;

use Msn::Util;
use Msn::Util::UserHash;

use IO::Select;
use Digest::MD5 qw/md5_hex/;

my %protocol_message_handler = (
    'VER' => sub { my ($bot, undef, @param) = @_;
        $bot->sendCommand("CVR", MSN_USER_AGENT, $bot->getUsername());
    },
    'CVR' => sub { my ($bot, undef, @param) = @_;
        $bot->sendCommand("USR", "TWN", "I", $bot->getUsername());
    },
    'INF' => sub { my ($bot, undef, @param) = @_;
    
    },
    'USR' => sub { my ($bot, undef, @param) = @_;
        if ($param[0] eq 'OK') {
            $bot->{client}->setUsername($param[1]);
            $bot->{client}->setNickname($param[2]);
            # TODO search - what is another param ???
            # $param[3] ?
            # $param[4] ?

            # TODO Add Login OK message handler.
            # $bot->{login_handler}->(@param);
            $bot->sendCommand('CHG', 'NLN');

            # If you did not send SYN command,
            # your status will be offline, but connection is active.
            if (grep /MSNP10/, MSNP_VERSIONS) {
                $bot->sendCommand('SYN', '0', '0');
            }
            else {
                $bot->sendCommand('SYN', 0);
            }
        }
        elsif ($param[0] eq 'TWN' and $param[1] eq 'S') {
            my $user = $bot->getUsername();
            my $pass = $bot->getPassword();
            my %param = split_param($param[-1]);
            my $ticket = Msn::Authentication::passport($user, $pass, \%param);

            $bot->sendCommand("USR", "TWN", "S", $ticket);
        }
        else {
            confess "Unsupported authentication method: $param[0]\n";
        }
    },
    'XFR' => sub { my ($bot, undef, $referralType, $address, @param) = @_;
        if ($referralType eq 'NS') {
        
            # Connection to notification server.
            my $connection = Msn::Connection->new($address);
            
            $bot->{connections}->add($connection->getId(), $connection);
            $bot->{select}->add($connection->getFilehandle());
            
            $bot->setCurrentSession($connection->getFilehandle());
            $bot->sendCommand("VER", MSNP_VERSIONS, "CVR0");
        }
        else {
            # XXX ... read protocol
            confess "NOT IMPLEMENTED: USR $referralType\n";
        }
    },
    'JOI' => sub { my ($bot, undef, @param) = @_;

        # TODO start session
        # 
        # my $msg = Msn::Message->new();
        # my $slp = MSN::SLP->new();
        # 
        # 
        # $msg->setBody($slp);
        # 
        # $bot->sendMessage($msg->dumpMessage());
        # 
    },
    'CAL' => sub { my ($bot, undef, @param) = @_;
    
    },
    'RNG' => sub { my ($bot, $TrID, $address, undef, $key, $username, $nickname) = @_;
        # When your friend clicked your name then ask you to join chat.
        
        # TODO Msn::Session object capcelize { socket, select }
        # 
        # - RNG
        # - XFR
        # - Msn::Bot::start()
        
        # TODO Connection's type "SB"
        
        my $connection = Msn::Connection->new($address);
        $bot->{connections}->add($connection->getId(), $connection);
        $bot->{select}->add($connection->getFilehandle());

        $bot->setCurrentSession($connection->getFilehandle());
        $bot->sendCommand("ANS", $bot->getUsername(), $key, $TrID);
    },
    'ANS' => sub { my ($bot, $TrID, @param) = @_;

        if (ref $bot->{answer_handler} eq 'CODE') {
            $bot->{answer_handler}->($bot, @param);
        }
        else {
            croak "answer handler not set\n";
        }
    },
    'MSG' => sub { my ($bot, $msgUsername, $msgNickname, $msgLength) = @_;
        my $connection = $bot->getCurrentSession();
        my $message = $connection->getMessage($msgLength);

        # TODO - capcelize switch MSG response by module or subroutine.

        my $recvMsg = Msn::Message->new(\$message);
        my $contentType = $recvMsg->getHeader('Content-Type');

        # TODO implement simple MIME parser or use MIME module.
        if ($contentType =~ m|^text\/plain\;\s*charset\=(.+)$|) {
            $bot->_call_handler("message_handler",
                                $msgUsername,
                                $msgNickname,
                                $recvMsg->getBody());
        }
        elsif ($contentType =~ m|^text/msmsgsprofile\b|) {
            # Hotmail
            print "MSG $msgNickname\n";
        }
        elsif ($contentType =~ m|^text/x-msmsgscontrol$|) {
            my $typingUser = $recvMsg->getHeader('TypingUser');
            print "MSG $msgNickname($typingUser) is typing\n";
        }
        elsif ($contentType =~ m|text/x-msmsgsinvite\b|) {
            # TODO - File Transfer
            my %param = map{/(.+)\: (.+)/} split(/\n/, $recvMsg->getBody());

            # TODO
            # ACCEPT field (Msn::Message::FileTransfer#accept)
            #
            # $bot->sendMessage(
            #  ...
            #  Msn::Message::FileTransfer->new(\%param)->accept();
            # )
            
            print "MSG - File Transfer \n";
        }
        elsif ($contentType =~ m|text/x-mms-emoticon\b|) {
            # TODO
            # take reference
            my $msnObject = Msn::Message::MsnObject->new($recvMsg->getBody());
            print qq(MSG emoticon @{[$msnObject->toString()]}\n);
        }
        elsif ($contentType =~ m|application/x-msnmsgrp2p$|) {
            # TODO read P2P packet.
            # my $p2p = Msn::Message::P2P->new($recvMsg->getBody());
        }
        else {
            # TODO
            print STDERR "Unsupport MSG content type\n";
            print STDERR "$contentType\n";
            print STDERR $recvMsg->dumpMessage()."\n";
        }
    },
    'LST' => sub { my ($bot, undef, $username, $nickname, @param) = @_;
        # TODO
        # Param of LSP is different format in MSNP10
    
        # TODO implement synchronize contact list.
        # TODO support GROUP param
        $bot->{client}->{contact_list}->add($username, $nickname);
    },
    'BPR' => sub { my ($bot, undef, @param) = @_;

    },
    'CHG' => sub { my ($bot, undef, @param) = @_;
    
    },
    'SYN' => sub {
        # Not recieve this command ?    
    },
    'GTC' => sub {

    },
    'BLP' => sub {

    },
    'LSG' => sub {
    
    },
    'IRO' => sub {
    
    },
    'QNG' => sub {
        # response of PNG
        # We have nothing to do.
    },
    'CHL' => sub { my ($bot, undef, $param) = @_;
        # server ping (Challenge)
        $bot->sendPing();

        # XXX
        # Client identification information
        my $clientStr = 'PROD0061VRRZH@4F';
        my $clientCode = 'JXQ6J@TUOGYV@N0M';
        
        $bot->sendQuery($clientStr, md5_hex($param.$clientCode));
    },
    'QRY' => sub {
        # do nothing
    },
    'NAK' => sub {

    },
    'ADC' => sub {

    },
    'PRP' => sub {

    },

    # TODO more command
    
    # notify status changed
    'NLN' => sub { my ($bot, undef, @param) = @_;
        # TODO
        # I'm not sure what command server notify to your bubby online.
        # $bot->_call_handler('status_handler', @param);
    },
    'ILN' => sub { my ($bot, undef, @param) = @_;

    },
    'FLN' => sub { my ($bot, undef, @param) = @_;

    },

    'BYE' => sub { my ($bot, undef, @param) = @_;

    },
);

sub new {
    my ($class, $username, $password) = @_;
    my $self = bless({}, $class);
    
    $self->{client} = Msn::Client->new($username, $password);
    $self->{select} = IO::Select->new();
    $self->{connections} = Msn::Util::UserHash->new();
    $self;
}

sub setMessageHandlers {
    my ($self, $param) = @_;
    
    $self->{status_handler}  = $param->{Status};
    $self->{message_handler} = $param->{Message};
    $self->{join_handler}    = $param->{Join};
    $self->{answer_handler}  = $param->{Answer};
    $self;
}

# setter/getter methods for current connection.
# set filehandle then get will return the Msn::Connection object
# by set Id. If you want to get filehandle for remove in IO::Select object.
#
# $connection = $bot->getCurrentSession();
# $filehandle = $connection->getFilehandle();
# $bot->{select}->remove($filehandle);
# $bot->{connections}->remove($filehandle->fileno());
#

sub setCurrentSession {
    my ($self, $filehandle) = @_;
    $self->{_current} = $filehandle;
}

sub getCurrentSession {
    my ($self) = @_;
    return $self->{connections}->get($self->{_current}->fileno());
}

# Wrapper methods
sub sendMessage {
    my ($self, $message) = @_;
    my $msg = Msn::Message->new();
	
    $msg->setHeader('MIME-Version', '1.0');
    $msg->setHeader('Content-Type', qq(text/plain; charset=UTF-8));
    $msg->setHeader('X-MMS-IM-Format', "FN=Arial; EF=; CO=0; CS=0; PF=22");
    $msg->setBody($message);
	
    $self->getCurrentSession()->sendMessage($msg->dumpMessage());
}

sub sendPing {
    my ($self) = @_;
    $self->getCurrentSession()->sendPing();
}

sub sendQuery {
    my ($self, $param, $query) = @_;
    $self->getCurrentSession()->sendQuery($param, $query);
}

sub sendCommand {
    my ($self) = shift(@_);
    $self->getCurrentSession()->sendCommand(@_);
}

sub getUsername {
    my ($self) = @_;
    return $self->{client}->getUsername();
}

sub getPassword {
    my ($self) = @_;
    return $self->{client}->getPassword();
}


sub run {
    my ($self) = @_;

    my $connection = Msn::Connection->new(MSN_DISPATCH_SERVER);
    
    $self->{select}->add($connection->getFilehandle());
    $self->{connections}->add($connection->getId(), $connection);
    
    $self->setCurrentSession($connection->getFilehandle());
    $self->sendCommand("VER", MSNP_VERSIONS, "CVR0");

    while (my @ready = $self->{select}->can_read()) {
        for my $filehandle (@ready) {
        
            $self->setCurrentSession($filehandle);
            my $conn = $self->getCurrentSession();
            my $line = $conn->getLine();
            
            # Disconnected
            unless (defined $line) {
                $self->{select}->remove($filehandle);
                $self->{connections}->remove($filehandle->fileno());
                next;
            }

            my ($cmd, $TrID, @param) = split(/\s+/, $line);
            if ($cmd =~ /^([0-9]{3})$/) {
                confess Msn::Error::getErrorMessage($1) . "\n";
            }
            elsif (ref $protocol_message_handler{$cmd} eq 'CODE') {
                $protocol_message_handler{$cmd}->($self, $TrID, @param);
            }
            else {
                print STDERR qq("$cmd" is not supported response.\n);
            }
        }
    }
}

sub start {
    my ($package, $username, $password) = @_;
    Msn::Bot->new($username, $password)->run();
}


sub _call_handler {
    my ($self, $handler_name, @args) = @_;

    $self->$handler_name(@args) if $self->can($handler_name);
    $self->{$handler_name}->($self, @args) if ref $self->{$handler_name} eq 'CODE';
}

sub status_handler {
    # override this method in subclass
}

sub message_handler {
    # override this method in subclass
}

sub join_handler {
    # override this method in subclass
}

sub answer_handler {
    # override this method in subclass
}

1;

