package Msn::Connection;

# $Id: Connection.pm 28 2004-03-20 04:56:41Z tea $

use strict;
use warnings;
use Carp;

use IO::Socket;

sub new($$) {
    my ($class, $uri) = @_;
    my $self = bless({}, $class);

print "Msn::Connection->new($uri)\n";

    $self->{connection} = IO::Socket::INET->new($uri);
    $self->{TrID} = 0;
    $self->{Debug} = 1;
    $self;
}

sub getFilehandle($) {
    my ($self) = @_;
    return $self->{connection};
}

sub getId($) {
    my ($self) = @_;
    return $self->{connection}->fileno();
}

sub sendRaw {
    my ($self, @args) = @_;
    
    my $msg = join(' ', @args);
    print ">>> @args" if $self->{Debug};
    
    $self->{connection}->send($msg);
}

sub sendCommand {
    my ($self, $command, @args) = @_;

    $self->sendRaw($command, $self->{TrID}++, @args, "\r\n");
}


# TODO
# Msn::Connection#sendPing
# Msn::Command#ping
sub sendPing {
    my ($self) = @_;
    $self->sendRaw("PNG", "\r\n");
}

sub sendMessage {
    my ($self, $msg) = @_;

    # If you send Message with wrong MSG header,
    # the server will disconnect client's connection,
    # WITHOUT return error code.
    # Be careful, "sendMessage" can not wrap "sendCommand"
    $self->sendRaw("MSG", $self->{TrID}++, "N", length($msg), "\r\n$msg");
}

# TODO
# Message and Query are "Payload" commands
sub sendQuery {
    my ($self, $param, $query) = @_;
    
    $self->sendRaw("QRY", $self->{TrID}++, $param, length($query), "\r\n$query");
}

sub getLine($) {
    my ($self) = @_;
    
    my $line = $self->{connection}->getline();

    return $line unless defined $line;
    chomp($line);

    print "<<< $line\r\n" if $self->{Debug};

    return $line;
}

sub getMessage($$) {
    my ($self, $msg_length) = @_;
    my $msg_body = "";

    $self->{connection}->read($msg_body, $msg_length);

    return $msg_body;
}

# !!! Warning !!!
# getline() will buffer read data.
# you can not use recv() after used getline function,
# without clean-up the read buffer.
sub recv {
    my ($self, $buffer_len) = @_;
    
    my $buffer;
    $self->{connection}->recv($buffer, $buffer_len||=1024);
    
    print "<<< $buffer\r\n" if $self->{Debug};
    
    return $buffer;
}

sub close($) {
    my ($self) = @_;
    $self->{connection}->close();
}

1;
