package Msn::Message::P2P;

# $Id: P2P.pm 32 2004-03-21 22:14:30Z tea $

use strict;
use warnings;
use Carp;

# XXX - Msn::Message::P2P has_a Msn::Message ???
# 
# Msn::Message is also subclass of PayloadCommand in modeling.
# 
# ------------  -----------------
# Msn::Message  Msn::Message::P2P
# ------------  -----------------
# - start line  - start line
# - header      - header:  Content-Type: application/x-msmsgrp2p
# - body        - body
# ------------      +-- header binary
#                   +-- -------------
#                   |   MSNSLP
#                   |   -------------
#                   |   - start line
#                   |   - header
#                   |   - body
#                   |   -------------
#                   +-- footer binary
#
#
# P2P's body contain MSNSLP packet.
#


# TODO
# I do not know this is good design or not.
# Let's me try ...
#
# use base qw/Msn::Message/;

# TODO
# use Msn::Util qw/randint/;

use Msn::Util::MIME;
use Msn::Message::MSNSLP;

sub new {
    my ($class, $destAddress) = @_;
    my $self = bless({}, $class);

    $self->{header} = Msn::Util::MIME->new();
    $self->{header}->setHeader('MIME-Version', '1.0');
    $self->{header}->setHeader('Content-Type', 'application/x-msnmsgrp2p');
    $self->{header}->setHeader('P2P-Dest', $destAddress);
    
    $self->{body} = Msn::Message::P2P::Chunk->new();

    $self;
}

sub makePacket {
    my ($self) = @_;
    my $body = $self->{body}->makePacket();
    my $header = "";

    $self->{header}->setHeader('Content-Length', length($body));
    $self->{header}->makeHeader();
    
    return join("\r\n", $header, $body);
}

# TODO rename Ack ?
package Msn::Message::P2P::Chunk;

# header
# 
#   48 byte binary  6DWORD and 3 QWORD are Big Endian order.
#   DWORD ... 4 byte (unsigned int)
#   QWORD ... 8 byte
#
#   [0] SessionID (DWORD)
#       which is zero when the clients are negotiating the session.
#       
#   [1] BaseID (DWORD)
#       random range from 4 to a max of 4294967295
#       
#   [2][3] Offset (QWORD)
#       only use when sent the data that larger than 1202 bytes.
#       For example: If you have some data which has a length 1496 bytes,
#       the first P2P message contains an offset field of 0,
#       the second contains an offset of 1202.
#       
#   [4][5] Total Data Size (QWORD)
#       
#   [6]
#
#   [7]
#
#   [8]
#
#   [9]
#
#
# body (optional data)
# 
#   see MSNSLP
#
#
# footer
# 
#   4 byte binary 1 DWORD(unsigned int) is Little Endian order.
#   and represent the ApplicationIndentifier(AppID).
#   If this field is zero, then the clients are negotiating the session.
#   The AppID is 0x1 for User Display Image and Emoticons,
#   0x2 for File Transfer and for Games, it's specified by the Start Menu code
# 
#   (4 byte, Little Encian order unsigned int)
#       0x1     User Display Image and Emoticons
#       0x2     File Transfer and Games
# 

sub new {
    my ($class, $header, $appId) = @_;
    my $self = bless({}, $class);

    # TODO - check assertion
    # assert(ref $header eq 'ARRAY');
    # assert(ref $footer eq 'ARRAY');

    $self->{header} = $header || [];
    $self->{appId}  = $appId || 0x0;
    $self->{data}   = Msn::Message::MSNSLP->new();

    return $self;
}

# TODO
# use Class:: module for generate thise getter/setter method.
# or make Msn::Util::Class
sub setSessionId {
    my ($self, $sid) = @_;
    $self->{header}->[0] = $sid;
}


sub makePacket {
    my ($self) = @_;

    my $data = $self->{data}->makePacket();

    $self->{header}->[ 0] ||= 0; # session Id
    $self->{header}->[ 1] ||= rand(4294967295) + 4; # TODO unique queue.
    $self->{header}->[ 2] ||= 0; # offset1
    $self->{header}->[ 3] ||= 0; # offset2
    $self->{header}->[ 4] ||= 0; # total size1
    $self->{header}->[ 5] ||= 0; # total size2
    $self->{header}->[ 6] ||= length($data); # message length
    $self->{header}->[ 7] ||= 0; # flag
    $self->{header}->[ 8] ||= 0; # ack session id
    $self->{header}->[ 9] ||= 0; # ack unique id
    $self->{header}->[10] ||= 0; # ack data size1
    $self->{header}->[11] ||= 0; # ack data size2
    
    my $header = pack("N12", @{$self->{header}}); # Big Endian order
    my $footer = pack("V", $self->{appId});       # Little Endian order

    return join('', $header,  $data, $footer);
}

1;

